/*
 * Copyright (C) 2007 Casey Schaufler <casey@schaufler-ca.com>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation, version 2.
 *
 * Author:
 *      Casey Schaufler <casey@schaufler-ca.com>
 *
 */

#include <linux/types.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/sched.h>
#include "smack.h"

static int printlog = 0;
static int root_cap = 0;

#define pr_smkdebug(fmt, ...)                      \
({                                             \
       if (printlog == 1)                      \
               printk(fmt, ##__VA_ARGS__);     \
       0;                                      \
})

static int __init params_setup(char *str)
{
       if (!str) {
               printlog = 0;
               return 0;
       }

       if (!strcmp("1", str)) {
               printlog = 1;
       } else {
               printk("Smack: cmdline smack.pr_debuglog=%s is not valid.", str);
               printlog = 0;
       }

       return 0;
}
__setup("smack.pr_debuglog=", params_setup);

static int __init cap_setup(char *str)
{
       if (!str) {
               root_cap = 0;
               return 0;
       }

       if (!strcmp("1", str)) {
               root_cap = 1;
       } else {
               printk("Smack: cmdline smack.root_cap=%s is not valid.", str);
               root_cap = 0;
       }

       return 0;
}
__setup("smack.root_cap=", cap_setup);

struct smack_known smack_known_huh = {
	.smk_known	= "?",
	.smk_secid	= 2,
};

struct smack_known smack_known_hat = {
	.smk_known	= "^",
	.smk_secid	= 3,
};

struct smack_known smack_known_star = {
	.smk_known	= "*",
	.smk_secid	= 4,
};

struct smack_known smack_known_floor = {
	.smk_known	= "_",
	.smk_secid	= 5,
};

struct smack_known smack_known_invalid = {
	.smk_known	= "",
	.smk_secid	= 6,
};

struct smack_known smack_known_web = {
	.smk_known	= "@",
	.smk_secid	= 7,
};

LIST_HEAD(smack_known_list);

/*
 * The initial value needs to be bigger than any of the
 * known values above.
 */
static u32 smack_next_secid = 10;

/*
 * what events do we log
 * can be overwritten at run-time by /smack/logging
 */
int log_policy = SMACK_AUDIT_DENIED;

/**
 * smk_access_entry - look up matching access rule
 * @subject_label: a pointer to the subject's Smack label
 * @object_label: a pointer to the object's Smack label
 * @rule_list: the list of rules to search
 *
 * This function looks up the subject/object pair in the
 * access rule list and returns the access mode. If no
 * entry is found returns -ENOENT.
 *
 * NOTE:
 *
 * Earlier versions of this function allowed for labels that
 * were not on the label list. This was done to allow for
 * labels to come over the network that had never been seen
 * before on this host. Unless the receiving socket has the
 * star label this will always result in a failure check. The
 * star labeled socket case is now handled in the networking
 * hooks so there is no case where the label is not on the
 * label list. Checking to see if the address of two labels
 * is the same is now a reliable test.
 *
 * Do the object check first because that is more
 * likely to differ.
 *
 * Allowing write access implies allowing locking.
 */
int smk_access_entry_orig(char *subject_label, char *object_label,
			struct list_head *rule_list)
{
	int may = -ENOENT;
	struct smack_rule *srp;

	list_for_each_entry_rcu(srp, rule_list, list) {
		if (srp->smk_object->smk_known == object_label &&
		    srp->smk_subject->smk_known == subject_label) {
			may = srp->smk_access;
			break;
		}
	}

	/*
	 * MAY_WRITE implies MAY_LOCK.
	 */
	if ((may & MAY_WRITE) == MAY_WRITE)
		may |= MAY_LOCK;
	return may;
}
/*New access_entry handle*/
/*只需要主体Label和客体Label的比较即可
 *客体对象需要区分文件Label。客体对象->如文件、消息队列、套接字、共享内存、信号量等，客体也可以是Linux进程或者IPC。客体对象主要是文件和进程
 *所以客体Label的长度是不固定的，客体Label举例：
 *网络socket Label "073b3a393845501d1e01" 10字节 20个字符： 权限级别 1字节  07，编号 4字节 3b3a3938，类别 4字节  45501d1e，自定义 1字节 01
 *文件Label 同上
 *主体Label 同上
 */
int smk_access_entry(char *subject_label, char *object_label, int request)
{
	int may = 0;  //由于0值已经被指定为未知错误，默认128是通过。128二进制：1000 0000, log太多，暂时改回0值
	int subj_lvl; //主体权限级别
	int subj_lvl_1;
	int subj_lvl_2;
	int obj_lvl; //客体权限级别
	int obj_lvl_1;
	int obj_lvl_2;

	int i;
	int index, bytelen, xgpos;

	unsigned char *buffer;
	unsigned char *obbuffer;
	unsigned char *subj_buffer;
	unsigned char *obj_buffer;
	int catfound = 0;
	int confound = 0;
	int cat;

	if(!smack_enabled_my)
		return 128;

	if(strlen(subject_label) > 5){
		subj_lvl_1=subject_label[0] - '0';
		subj_lvl_2=subject_label[1] - '0';
		subj_lvl=subj_lvl_1*16+subj_lvl_2;
	}else{
		subj_lvl=0;
	}
	if(strlen(object_label) > 5){ //客体为进程
		obj_lvl_1=object_label[0] - '0';
		obj_lvl_2=object_label[1] - '0';
		obj_lvl=obj_lvl_1*16+obj_lvl_2;
	}else{
	    obj_lvl=0; //客体Label _、*、等
	}
	//权限级别管控

	if(strlen(subject_label) < 5)
		return 0; /*如果主体Label为_等默认Label，则PASS，返回128.*/
	if(subj_lvl < obj_lvl){
		may = 1; // Denied   0000 0001
		pr_smkdebug("-------------------MJ denied. subj-lvl [%d] < obj_lvl [%d].\n", subj_lvl, obj_lvl);//调试信息
	}

	//类别管控
	if(strlen(object_label) > 5  || strlen(object_label) > 5){
		bytelen = strlen(subject_label) / 2; // 字符串长度的一半

		buffer = kzalloc(bytelen, GFP_KERNEL);
		if (buffer == NULL)
		    return ERR_PTR(-ENOMEM);
		memset(buffer,0,bytelen);

		obbuffer = kzalloc(14, GFP_KERNEL);
		if (obbuffer == NULL)
		    return ERR_PTR(-ENOMEM);
		memset(obbuffer,0,14);

		/*转换为字节码  object*/
		for(i= 0; i < 14; i++){
		    index = i * 2;
		    obbuffer[i] = ((toByteXG(object_label[index]))<<4) | toByteXG(object_label[index+1]);
		}
		/*Object cat and con*/
		int obj_cat = (obbuffer[5] << 24 | obbuffer[6] << 16 | obbuffer[7] << 8 | obbuffer[8]);


		/*转换为字节码 subject*/
		for(i= 0; i < bytelen; i++){
		    index = i * 2;
		    buffer[i] = ((toByteXG(subject_label[index]))<<4) | toByteXG(subject_label[index+1]);
		}

		cat = (buffer[5] << 24 | buffer[6] << 16 | buffer[7] << 8 | buffer[8]);
		if(obj_cat == cat)
			catfound = 1;

		if(!catfound){
		    may = may | 4; /*Return error no 0000 0100*/
		}

		if(may == 0){
			return 128;  //只有文件Label的管控放行时才return 128.
		}

		kfree(buffer);
		kfree(obbuffer);

	}
	/*
	 * MAY_WRITE implies MAY_LOCK.
	 */
	if(may > 0){
	    return may;
	}
	if(may == 0){
	    return 0;
	}
	if(may < 0){
	    return may;
	}

	return 0;
}

/*
 * smk_access - determine if a subject has a specific access to an object
 * @subject: a pointer to the subject's Smack label entry
 * @object: a pointer to the object's Smack label entry
 * @request: the access requested, in "MAY" format
 * @a : a pointer to the audit data
 *
 * This function looks up the subject/object pair in the
*/
int smk_access(struct smack_known *subject, struct smack_known *object,
	       int request, struct smk_audit_info *a)
{
	int may = MAY_NOT;
	int rc = 0;

	if(!smack_enabled_my)
		return 0;

	/*
	 * Hardcoded comparisons.
	 */
	/*
	 * A star subject can't access any object.
	 */
	if (subject == &smack_known_star) {
		rc = -EACCES;
		goto out_audit;
	}
	/*
	 * An internet object can be accessed by any subject.
	 * Tasks cannot be assigned the internet label.
	 * An internet subject can access any object.
	 */
	////if (object == &smack_known_web || subject == &smack_known_web)
	////	goto out_audit;
	/*
	 * A star object can be accessed by any subject.
	 */
	////if (object == &smack_known_star)
	////	goto out_audit;
	/*
	 * An object can be accessed in any way by a subject
	 * with the same label.
	 */
	////if (subject->smk_known == object->smk_known)
	////	goto out_audit;
	/*
	 * A hat subject can read or lock any object.
	 * A floor object can be read or locked by any subject.
	 */
	/*if ((request & MAY_ANYREAD) == request ||
	    (request & MAY_LOCK) == request) {
		if (object == &smack_known_floor)
			goto out_audit;
		if (subject == &smack_known_hat)
			goto out_audit;
	}*/
	/*
	 * Beyond here an explicit relationship is required.
	 * If the requested access is contained in the available
	 * access (e.g. read is included in readwrite) it's
	 * good. A negative response from smk_access_entry()
	 * indicates there is no entry for this pair.
	 */
	/*rcu_read_lock();
	may = smk_access_entry(subject->smk_known, object->smk_known,
			       &subject->smk_rules);
	rcu_read_unlock();*/ //原函数操作，现在需要跳过这些操作，后面可能用到这些操作，先暂时注释
	may = smk_access_entry(subject->smk_known, object->smk_known, request);

	if (may < 0) {
		rc = -EACCES;
	}else{
	   if (may == 128 || may == 0)
	      rc = 0;
	   else
	      rc = -EACCES;
	}
	/*if (may <= 0 || (request & may) != request) {
		rc = -EACCES;
		goto out_audit;
	}*/ //暂时取消audit功能
#ifdef CONFIG_SECURITY_SMACK_BRINGUP
	/*
	 * Return a positive value if using bringup mode.
	 * This allows the hooks to identify checks that
	 * succeed because of "b" rules.
	 */
	//if (may & MAY_BRINGUP)
	//	rc = SMACK_BRINGUP_ALLOW;//现在不需要的操作
#endif

out_audit:

#ifdef CONFIG_SECURITY_SMACK_BRINGUP
/*	if (rc < 0) {
		if (object == smack_unconfined)
			rc = SMACK_UNCONFINED_OBJECT;
		if (subject == smack_unconfined)
			rc = SMACK_UNCONFINED_SUBJECT;
	}
*/ //现在不需要的操作
#endif

#ifdef CONFIG_AUDIT
if(may > 0 && may != 128){
//if(rc > 0){//原函数操作
	if (a)
		smackxg_log(subject->smk_known, object->smk_known,
			  request, may, a);
			  //request, rc, a);
	else
		pr_smkdebug("audit 1 failed\n");//调试信息
}
if(rc < 0){
//if(strcmp(current->comm, "systemd-journal") != 0 && strcmp(current->comm, "auditd") != 0 && strcmp(current->comm, "Xorg") != 0 && strcmp(current->comm, "cdos-desktop") != 0 && strcmp(current->comm, "radeon_cs:0") != 0){//定制打印的log信息，现不再使用
	if (a)
		smackxg_log(subject->smk_known, object->smk_known,
			  request, may, a);
	//else
		//printk("audit 2 failed\n");//调试信息
}

#endif

	return rc;
}



/**
 * smk_access - determine if a subject has a specific access to an object
 * @subject: a pointer to the subject's Smack label entry
 * @object: a pointer to the object's Smack label entry
 * @request: the access requested, in "MAY" format
 * @a : a pointer to the audit data
 *
 * This function looks up the subject/object pair in the
 * access rule list and returns 0 if the access is permitted,
 * non zero otherwise.
 *
 * Smack labels are shared on smack_list
 */
int smk_access_orig(struct smack_known *subject, struct smack_known *object,
	       int request, struct smk_audit_info *a)
{
	int may = MAY_NOT;
	int rc = 0;

	/*
	 * Hardcoded comparisons.
	 */
	/*
	 * A star subject can't access any object.
	 */
	if (subject == &smack_known_star) {
		rc = -EACCES;
		goto out_audit;
	}
	/*
	 * An internet object can be accessed by any subject.
	 * Tasks cannot be assigned the internet label.
	 * An internet subject can access any object.
	 */
	if (object == &smack_known_web || subject == &smack_known_web)
		goto out_audit;
	/*
	 * A star object can be accessed by any subject.
	 */
	if (object == &smack_known_star)
		goto out_audit;
	/*
	 * An object can be accessed in any way by a subject
	 * with the same label.
	 */
	if (subject->smk_known == object->smk_known)
		goto out_audit;
	/*
	 * A hat subject can read or lock any object.
	 * A floor object can be read or locked by any subject.
	 */
	if ((request & MAY_ANYREAD) == request ||
	    (request & MAY_LOCK) == request) {
		if (object == &smack_known_floor)
			goto out_audit;
		if (subject == &smack_known_hat)
			goto out_audit;
	}
	/*
	 * Beyond here an explicit relationship is required.
	 * If the requested access is contained in the available
	 * access (e.g. read is included in readwrite) it's
	 * good. A negative response from smk_access_entry()
	 * indicates there is no entry for this pair.
	 */
	rcu_read_lock();
	may = smk_access_entry(subject->smk_known, object->smk_known,
			       &subject->smk_rules);
	rcu_read_unlock();

	if (may <= 0 || (request & may) != request) {
		rc = -EACCES;
		goto out_audit;
	}
#ifdef CONFIG_SECURITY_SMACK_BRINGUP
	/*
	 * Return a positive value if using bringup mode.
	 * This allows the hooks to identify checks that
	 * succeed because of "b" rules.
	 */
	if (may & MAY_BRINGUP)
		rc = SMACK_BRINGUP_ALLOW;
#endif

out_audit:

#ifdef CONFIG_SECURITY_SMACK_BRINGUP
	if (rc < 0) {
		if (object == smack_unconfined)
			rc = SMACK_UNCONFINED_OBJECT;
		if (subject == smack_unconfined)
			rc = SMACK_UNCONFINED_SUBJECT;
	}
#endif

#ifdef CONFIG_AUDIT
	if (a)
		smack_log(subject->smk_known, object->smk_known,
			  request, rc, a);
#endif

	return rc;
}

/**
 * smk_tskacc - determine if a task has a specific access to an object
 * @tsp: a pointer to the subject's task
 * @obj_known: a pointer to the object's label entry
 * @mode: the access requested, in "MAY" format
 * @a : common audit data
 *
 * This function checks the subject task's label/object label pair
 * in the access rule list and returns 0 if the access is permitted,
 * non zero otherwise. It allows that the task may have the capability
 * to override the rules.
 */
int smk_tskacc(struct task_smack *tsp, struct smack_known *obj_known,
	       u32 mode, struct smk_audit_info *a)
{
	struct smack_known *sbj_known = smk_of_task(tsp);
	int may;
	int rc;
	rc = 0;
        if (sbj_known == &smack_known_star) {
		rc = -EACCES;
		goto out_audit;
	}
	/*
	 * Check the global rule list
	 */
	rc = smk_access(sbj_known, obj_known, mode, a);
		/*
		 * If there is an entry in the task's rule list
		 * it can further restrict access.
		 */
	////	may = smk_access_entry(sbj_known->smk_known,
	////			       obj_known->smk_known,
	////			       mode);
				       //&tsp->smk_rules);
	////	if (may < 0)
	////		goto out_audit;
	////	if ((mode & may) == mode)
	////		goto out_audit;
	////	rc = -EACCES;
	////}  //原函数操作，需要跳过

	/*
	 * Allow for priviliged to override policy.
	 */
	if (root_cap == 1){
	    if (rc != 0 && smack_privileged(CAP_MAC_OVERRIDE)){
		rc = 0;
		pr_smkdebug("--------------------> privileged \n");
	    }
	}
out_audit:
#ifdef CONFIG_AUDIT
/*
	if (a)
		smack_log(sbj_known->smk_known, obj_known->smk_known,
			  mode, rc, a);
*/ //取消audit的功能
#endif
	return rc;
}

/**
 * smk_curacc - determine if current has a specific access to an object
 * @obj_known: a pointer to the object's Smack label entry
 * @mode: the access requested, in "MAY" format
 * @a : common audit data
 *
 * This function checks the current subject label/object label pair
 * in the access rule list and returns 0 if the access is permitted,
 * non zero otherwise. It allows that current may have the capability
 * to override the rules.
 */
int smk_curacc(struct smack_known *obj_known,
	       u32 mode, struct smk_audit_info *a)
{
	struct task_smack *tsp = current_security();

	return smk_tskacc(tsp, obj_known, mode, a);
}

#ifdef CONFIG_AUDIT
/**
 * smack_str_from_perm : helper to transalate an int to a
 * readable string
 * @string : the string to fill
 * @access : the int
 *
 */
static inline void smack_str_from_perm(char *string, int access)
{
	int i = 0;

	if (access & MAY_READ)
		string[i++] = 'r';
	if (access & MAY_WRITE)
		string[i++] = 'w';
	if (access & MAY_EXEC)
		string[i++] = 'x';
	if (access & MAY_APPEND)
		string[i++] = 'a';
	if (access & MAY_TRANSMUTE)
		string[i++] = 't';
	if (access & MAY_LOCK)
		string[i++] = 'l';
	string[i] = '\0';
}
/**
 * smack_log_callback - SMACK specific information
 * will be called by generic audit code
 * @ab : the audit_buffer
 * @a  : audit_data
 *
 */
static void smack_log_callback(struct audit_buffer *ab, void *a)
{
	struct common_audit_data *ad = a;
	struct smack_audit_data *sad = ad->smack_audit_data;
	audit_log_format(ab, "lsm=SMACK fn=%s action=%s",
			 ad->smack_audit_data->function,
			 sad->result ? "denied" : "granted");
	audit_log_format(ab, " subject=");
	audit_log_untrustedstring(ab, sad->subject);
	audit_log_format(ab, " object=");
	audit_log_untrustedstring(ab, sad->object);
	if (sad->request[0] == '\0')
		audit_log_format(ab, " labels_differ");
	else
		audit_log_format(ab, " requested=%s", sad->request);
}

void audit_d_path(char *p, const struct path *path)
{
	char *pathname;
	/* We will allow 11 spaces for ' (deleted)' to be appended */
	pathname = kmalloc(PATH_MAX+11, GFP_NOFS);
	if (!pathname) {
	    return;
	}
	p = d_path(path, pathname, PATH_MAX+11);
	kfree(pathname);
}
/* New smack_log
 *
 */
void smackxg_log(char *subject_label, char *object_label, int request,
	       int result, struct smk_audit_info *ad)
{
	char logbuffer[512] = {0}; /*log最长字符，并初始化0*/
	char logbuffer2[512] = {0}; /*log最长字符，并初始化0*/
	char tmpbuf[128] = {0}; /*临时字符串*/
	char comm[sizeof(current->comm)];

#ifdef CONFIG_SECURITY_SMACK_BRINGUP
	char request_buffer[SMK_NUM_ACCESS_TYPE + 5];
#else
	char request_buffer[SMK_NUM_ACCESS_TYPE + 1];
#endif
	struct smack_audit_data *sad;
	struct common_audit_data *a = &ad->a;
	struct inode *inode;

	/*if (result < 0 && (log_policy & SMACK_AUDIT_DENIED) == 0)
		return;
	if (result == 0 && (log_policy & SMACK_AUDIT_ACCEPT) == 0)
		return;
	*/
	BUILD_BUG_ON(sizeof(a->u) > sizeof(void *)*2);

	memcpy(comm, current->comm, sizeof(comm));

	sad = a->smack_audit_data;
	if (sad->function == NULL)
		sad->function = "unknown";
	/* end preparing the audit data */
	smack_str_from_perm(request_buffer, request);
	sad->request = request_buffer;
	sad->subject = subject_label;
	sad->object = object_label;

	sprintf(tmpbuf, "fn=%s", sad->function);
	strcat(logbuffer2, tmpbuf);

	if(current->loginuid.val == 4294967295 || current->loginuid.val ==0)
	    sprintf(tmpbuf, "user=[%s]        ", "root");
	else
	    sprintf(tmpbuf, "userid=[%d]        ", current->loginuid.val);
	if(strlen(tmpbuf)>0)
	    strcat(logbuffer, tmpbuf);
	else
	    strcat(logbuffer, "user=[null]        ");
	memset(tmpbuf, 0, sizeof(tmpbuf));
	//Action: r,w,x,...
	sprintf(tmpbuf, "action=[%s]        ", sad->request);
	if(strlen(tmpbuf)>0)
	    strcat(logbuffer, tmpbuf);
	else
	    strcat(logbuffer, "action=[null]        ");
	memset(tmpbuf, 0, sizeof(tmpbuf));

	//common Name
	sprintf(tmpbuf, "common=[%s]        ", comm);
	if(strlen(tmpbuf)>0)
	    strcat(logbuffer, tmpbuf);
	else
	    strcat(logbuffer, "common=[null]        ");
	memset(tmpbuf, 0, sizeof(tmpbuf));
	//Subject Label
	sprintf(tmpbuf, " subject_label=[%s]        ", sad->subject);
	if(strlen(tmpbuf)>0)
	    strcat(logbuffer, tmpbuf);
	else
	    strcat(logbuffer, "subject_label=[null]        ");
	memset(tmpbuf, 0, sizeof(tmpbuf));

	//Object path
	switch (a->type) {
	case LSM_AUDIT_DATA_PATH:{
	    char *pathname, *p;
	    //pathname = kmalloc(PATH_MAX+11, GFP_NOFS);
	    pathname = kmalloc(PATH_MAX+11, GFP_KERNEL);
	    if (!pathname) {
	        return;
	    }
	    memset(pathname, '\0', PATH_MAX+11);
	    p = d_path(&a->u.path, pathname, PATH_MAX+11);
	    pr_smkdebug("full path ---------------p [%s] a->type [%s]\n", p, a->type);//调试信息
	    sprintf(tmpbuf, "object_path=[%s]        ", p);

	    kfree(pathname);
	    inode = d_backing_inode(a->u.path.dentry);
	    }
	    break;
	case LSM_AUDIT_DATA_FILE: {
	    char *pathname, *p;
	    pathname = kmalloc(PATH_MAX+11, GFP_KERNEL);
	    if (!pathname) {
	        return;
	    }
	    memset(pathname, '\0', PATH_MAX+11);
	    p = d_path(&a->u.file->f_path, pathname, PATH_MAX+11);
	    pr_smkdebug("full file path---------------p [%s] a->type [%s]\n", p, a->type);//调试信息
	    sprintf(tmpbuf, "object_path=[%s]        ", p);

	    inode = file_inode(a->u.file);
	    }
	    break;
	case LSM_AUDIT_DATA_DENTRY: {
	    inode = d_backing_inode(a->u.dentry);
	    sprintf(tmpbuf, "object_name=[%s]        ", a->u.dentry->d_name.name);
	    }
	    break;
	case LSM_AUDIT_DATA_IOCTL_OP: {
	    char *pathname, *p;
	    pathname = kmalloc(PATH_MAX+11, GFP_KERNEL);
	    if (!pathname) {
	        return;
	    }
	    memset(pathname, '\0', PATH_MAX+11);
	    p = d_path(&a->u.op->path, pathname, PATH_MAX+11);
	    pr_smkdebug("full data ioctl  ---------------p [%s] a->type [%s]\n", p, a->type);//调试信息
	    sprintf(tmpbuf, "object_path=[%s]        ", p);

	    inode = a->u.op->path.dentry->d_inode;
	    }
	    break;

	case LSM_AUDIT_DATA_INODE: {
	    struct dentry *dentry;
	    inode = a->u.inode;
	    dentry = d_find_alias(inode);
	    if(dentry)
	        sprintf(tmpbuf, "object_name=[%s]        ", dentry->d_name.name);
	    }
	    break;

	/*case LSM_AUDIT_DATA_TASK: {
		struct task_struct *tsk = a->u.tsk;
		if (tsk) {
			pid_t pid = task_tgid_nr(tsk);
			if (pid) {
				char comm[sizeof(tsk->comm)];
				sprintf(tmpbuf, " objecttask=%s", comm);
			}
		}
		break;
	}*/ //TASK审计信息不再需要

	}
	if(strlen(tmpbuf)>0)
	    strcat(logbuffer, tmpbuf);
	else
	    strcat(logbuffer, "null        ");
	//strcat(logbuffer, tmpbuf);
	memset(tmpbuf, 0, sizeof(tmpbuf));

	//Object Label
	sprintf(tmpbuf, "object_label=[%s]        ", object_label);
	if(strlen(tmpbuf)>0)
	    strcat(logbuffer, tmpbuf);
	else
	    strcat(logbuffer, "null        ");
	memset(tmpbuf, 0, sizeof(tmpbuf));
	//Result
	if(result == 0 || result == 128) //只有result=0 & 128时才是pass
	    strcat(logbuffer, "result=[success]        ");
	else
	    strcat(logbuffer, "result=[fail]        ");

	//Errorno
	if(result > 128)
	    sprintf(tmpbuf, "error_num=[%d]", result - 128);
	else
	    sprintf(tmpbuf, "error_num=[%d]", result);
	if(strlen(tmpbuf)>0)
	    strcat(logbuffer, tmpbuf);
	else
	    strcat(logbuffer, "null        ");
	memset(tmpbuf, 0, sizeof(tmpbuf));

	sprintf(tmpbuf, " requested=%s pid=%d comm=%s ", sad->request, task_tgid_nr(current), comm);
	strcat(logbuffer2, tmpbuf);
	memset(tmpbuf, 0, sizeof(tmpbuf));


	//sprintf(tmpbuf, " requested=%s pid=%d comm=%s ", sad->request, task_tgid_nr(current), comm);//调试信息

        /*if (inode) {
	    memset(tmpbuf, 0, sizeof(tmpbuf));
	    sprintf(tmpbuf, " dev=%s ino=%lu", inode->i_sb->s_id, inode->i_ino);
	    strcat(logbuffer, tmpbuf);
        }*/  //调试信息

	//屏蔽kernel swapper/1, swapper/2, in:imklog进程的log
	if(strcmp(comm, "swapper") > 0 || strcmp(comm, "in:imklog") == 0){
	    return;
	}else{
	    pr_info("accesslog: %s \n", logbuffer);
	    pr_smkdebug(KERN_ERR "%s[%d] accesslog: devid    %s \n",  comm, task_tgid_nr(current), logbuffer);//调试信息
	    pr_info("      log: [%s] v-result [%d] a->type[%x]\n", logbuffer2, result, a->type);
	}

}

/**
 *  smack_log - Audit the granting or denial of permissions.
 *  @subject_label : smack label of the requester
 *  @object_label  : smack label of the object being accessed
 *  @request: requested permissions
 *  @result: result from smk_access
 *  @a:  auxiliary audit data
 *
 * Audit the granting or denial of permissions in accordance
 * with the policy.
 */
void smack_log(char *subject_label, char *object_label, int request,
	       int result, struct smk_audit_info *ad)
{
#ifdef CONFIG_SECURITY_SMACK_BRINGUP
	char request_buffer[SMK_NUM_ACCESS_TYPE + 5];
#else
	char request_buffer[SMK_NUM_ACCESS_TYPE + 1];
#endif
	struct smack_audit_data *sad;
	struct common_audit_data *a = &ad->a;

	/* check if we have to log the current event */
	/*if (result < 0 && (log_policy & SMACK_AUDIT_DENIED) == 0)
		return;
	if (result == 0 && (log_policy & SMACK_AUDIT_ACCEPT) == 0)
		return;
	*/ //调试节点
	sad = a->smack_audit_data;

	if (sad->function == NULL)
		sad->function = "unknown";

	/* end preparing the audit data */
	smack_str_from_perm(request_buffer, request);
	sad->subject = subject_label;
	sad->object  = object_label;
#ifdef CONFIG_SECURITY_SMACK_BRINGUP
	/*
	 * The result may be positive in bringup mode.
	 * A positive result is an allow, but not for normal reasons.
	 * Mark it as successful, but don't filter it out even if
	 * the logging policy says to do so.
	 */
	if (result == SMACK_UNCONFINED_SUBJECT)
		strcat(request_buffer, "(US)");
	else if (result == SMACK_UNCONFINED_OBJECT)
		strcat(request_buffer, "(UO)");

	if (result > 0)
		result = 0;
#endif
	sad->request = request_buffer;
	sad->result  = result;

	common_lsm_audit(a, smack_log_callback, NULL);
}
#else /* #ifdef CONFIG_AUDIT */
void smack_log(char *subject_label, char *object_label, int request,
               int result, struct smk_audit_info *ad)
{
}
#endif

DEFINE_MUTEX(smack_known_lock);

struct hlist_head smack_known_hash[SMACK_HASH_SLOTS];

/**
 * smk_insert_entry - insert a smack label into a hash map,
 *
 * this function must be called under smack_known_lock
 */
void smk_insert_entry(struct smack_known *skp)
{
	unsigned int hash;
	struct hlist_head *head;

	hash = full_name_hash(NULL, skp->smk_known, strlen(skp->smk_known));
	head = &smack_known_hash[hash & (SMACK_HASH_SLOTS - 1)];

	hlist_add_head_rcu(&skp->smk_hashed, head);
	list_add_rcu(&skp->list, &smack_known_list);
}

/**
 * smk_find_entry - find a label on the list, return the list entry
 * @string: a text string that might be a Smack label
 *
 * Returns a pointer to the entry in the label list that
 * matches the passed string or NULL if not found.
 */
struct smack_known *smk_find_entry(const char *string)
{
	unsigned int hash;
	struct hlist_head *head;
	struct smack_known *skp;

	hash = full_name_hash(NULL, string, strlen(string));
	head = &smack_known_hash[hash & (SMACK_HASH_SLOTS - 1)];

	hlist_for_each_entry_rcu(skp, head, smk_hashed)
		if (strcmp(skp->smk_known, string) == 0)
			return skp;

	return NULL;
}

/**
 * secmack_parser2 - parse a user's secmark New one
 * @secstr: a text string of secmark
 * Returns a pointer to the entry commonSM
 */

static int secmack_parser2(const char *secstr, struct subSecmark *xgattr)
{
	int i;
	int index, bytelen, xgpos;
	unsigned char *buffer;

        buffer = kzalloc(SMK_LONGLABEL, GFP_KERNEL);
        if (buffer == NULL)
            return ERR_PTR(-ENOMEM);
	memset(buffer,0,strlen(buffer));
	bytelen = strlen(secstr) / 2; // 字符串长度的一半
	/*转换为字节码*/
	for(i= 0; i < bytelen; i++){
	    index = i * 2;
	    buffer[i] = ((toByteXG(secstr[index]))<<4) | toByteXG(secstr[index+1]);
	}
        xgattr->level = buffer[0] - 0;  // 权限级别 2个char 07
	xgattr->semanteme = (buffer[1] << 24 | buffer[2] << 16 | buffer[3] << 8 | buffer[4]);  // 编号 8个char
	xgattr->cat = (buffer[5] << 24 | buffer[6] << 16 | buffer[7] << 8 | buffer[8]);  // 类别 8个char
	xgattr->self_defined = buffer[9];//自定义 2个char

	pr_smkdebug("nic_level :[%0x] nic_semanteme: [%04x] nic_cat = [%04x] nic_self_defined = [%0x]\n",
		xgattr->level,xgattr->semanteme, xgattr->cat, xgattr->self_defined);	       //调试信息
	kfree(buffer);
	return 1;
}

/**
 * secmack_parser - parse a user's secmark
 * @secstr: a text string of secmark
 * Returns a pointer to the entry commonSM
 */
static struct secmark *secmack_parser(const char *secstr)
{
	struct secmark *xgattr;
	int i;
	int index, bytelen, xgpos;
	unsigned char *buffer;

	xgattr = kzalloc(sizeof(*xgattr), GFP_KERNEL);
	xgattr->commonSM = kzalloc(sizeof(*xgattr->commonSM), GFP_KERNEL);
        buffer = kzalloc(SMK_LONGLABEL, GFP_KERNEL);
        if (buffer == NULL)
            return ERR_PTR(-ENOMEM);
	memset(buffer,0,strlen(buffer));
	bytelen = strlen(secstr) / 2; // 字符串长度的一半
	/*转换为字节码*/
	for(i= 0; i < bytelen; i++){
	    index = i * 2;
	    buffer[i] = ((toByteXG(secstr[index]))<<4) | toByteXG(secstr[index+1]);
	}
	xgattr->commonSM->level = buffer[0] - 0;  // 权限级别 2个char 07
	xgattr->commonSM->semanteme = (buffer[1] << 24 | buffer[2] << 16 | buffer[3] << 8 | buffer[4]);  // 编号 8个char
	xgattr->commonSM->cat = (buffer[5] << 24 | buffer[6] << 16 | buffer[7] << 8 | buffer[8]);  // 类别 8个char
	xgattr->commonSM->self_defined = buffer[9];//自定义 2个char

	pr_smkdebug("nic_level :[%0x] nic_semanteme: [%04x] nic_cat = [%04x] nic_self_defined = [%0x]\n",
		xgattr->commonSM->level,xgattr->commonSM->semanteme, xgattr->commonSM->cat, xgattr->commonSM->self_defined);//调试信息
	kfree(buffer);
	return xgattr;
}

/**
 * smk_parse_smack - parse smack label from a text string
 * @string: a text string that might contain a Smack label
 * @len: the maximum size, or zero if it is NULL terminated.
 *
 * Returns a pointer to the clean label or an error code.
 */
char *smk_parse_smack(const char *string, int len)
{
	char *smack;
	int i;

	if (len <= 0)
		len = strlen(string) + 1;

	/*
	 * Reserve a leading '-' as an indicator that
	 * this isn't a label, but an option to interfaces
	 * including /smack/cipso and /smack/cipso2
	 */
	if (string[0] == '-')
		return ERR_PTR(-EINVAL);

	for (i = 0; i < len; i++)
		if (string[i] > '~' || string[i] <= ' ' || string[i] == '/' ||
		    string[i] == '"' || string[i] == '\\' || string[i] == '\'')
			break;

	if (i == 0 || i >= SMK_LONGLABEL)
		return ERR_PTR(-EINVAL);

	smack = kzalloc(i + 1, GFP_NOFS);
	if (smack == NULL)
		return ERR_PTR(-ENOMEM);

	strncpy(smack, string, i);

	return smack;
}

/**
 * smk_netlbl_mls - convert a catset to netlabel mls categories
 * @catset: the Smack categories
 * @sap: where to put the netlabel categories
 *
 * Allocates and fills attr.mls
 * Returns 0 on success, error code on failure.
 */
int smk_netlbl_mls(int level, char *catset, struct netlbl_lsm_secattr *sap,
			int len)
{
	unsigned char *cp;
	unsigned char m;
	int cat;
	int rc;
	int byte;

	sap->flags |= NETLBL_SECATTR_MLS_CAT;
	sap->attr.mls.lvl = level;
	sap->attr.mls.cat = NULL;

	for (cat = 1, cp = catset, byte = 0; byte < len; cp++, byte++)
		for (m = 0x80; m != 0; m >>= 1, cat++) {
			if ((m & *cp) == 0)
				continue;
			rc = netlbl_catmap_setbit(&sap->attr.mls.cat,
						  cat, GFP_ATOMIC);
			if (rc < 0) {
				netlbl_catmap_free(sap->attr.mls.cat);
				return rc;
			}
		}

	return 0;
}

/**
 * smk_import_entry - import a label, return the list entry
 * @string: a text string that might be a Smack label
 * @len: the maximum size, or zero if it is NULL terminated.
 *
 * Returns a pointer to the entry in the label list that
 * matches the passed string, adding it if necessary,
 * or an error code.
 */
struct smack_known *smk_import_entry(const char *string, int len)
{
	struct smack_known *skp;
	char *smack;
	int slen;
	int rc,res;
	//char *leftstr;
	char mjstr[64] =  {'\0'};
	char doistr[16] =  {""};
	char catstr[32] =  {""};
	//char maincatstr[48] =  {""};
	int smlen;
	int flag = 0;
	struct secmark *common_sec = NULL;
	char temp[32] = {""};
	int typehigh;

	smack = smk_parse_smack(string, len);
	if (IS_ERR(smack))
		return ERR_CAST(smack);

	mutex_lock(&smack_known_lock);

	skp = smk_find_entry(smack);
	if (skp != NULL)
		goto freeout;

	skp = kzalloc(sizeof(*skp), GFP_NOFS);
	if (skp == NULL) {
		skp = ERR_PTR(-ENOMEM);
		goto freeout;
	}

	skp->smk_known = smack;
	skp->smk_secid = smack_next_secid++;

	memset(mjstr, 0, sizeof(mjstr));
	memset(doistr, 0, sizeof(doistr));
	memset(catstr, 0, sizeof(catstr));
	memset(temp, 0, sizeof(temp));
	slen = strlen(smack);
	//当len=0时，是导入用户M标。
	smlen = slen - 18;
        pr_smkdebug("slen [%d] len [%d] smack [%s] smlen [%d]\n", slen, len, smack, smlen);//调试信息
	if(smlen == 2){//all label len are 20

	    sprintf(temp, "%1x%1x", toByteXG(smack[0]),toByteXG(smack[1])); //权限级别
	    strcat(mjstr, temp);
	    sprintf(temp, "%1x%1x%1x%1x%1x%1x%1x%1x", toByteXG(smack[2]),toByteXG(smack[3]),toByteXG(smack[4]),toByteXG(smack[5]),toByteXG(smack[6]),toByteXG(smack[7]),toByteXG(smack[8]),toByteXG(smack[9])); //编号
	    strcat(mjstr, temp);
	    sprintf(temp, "%1x%1x%1x%1x%1x%1x%1x%1x", toByteXG(smack[10]),toByteXG(smack[11]),toByteXG(smack[12]),toByteXG(smack[13]),toByteXG(smack[14]),toByteXG(smack[15]),toByteXG(smack[16]),toByteXG(smack[17])); //类别
	    strcat(mjstr, temp);
	    sprintf(temp, "%1x%1x", toByteXG(smack[18]),toByteXG(smack[19])); //自定义
	    strcat(mjstr, temp);

            pr_smkdebug("F-mjstr [%s] strlen-mjstr [%lu] smack [%s] . \n", mjstr, strlen(mjstr), smack);//调试信息
	    mjstr[20]='\0';
	    skp->smk_netlabel.domain = kzalloc(strlen(mjstr) + 1, GFP_KERNEL);
	    memset(skp->smk_netlabel.domain, 0, strlen(mjstr) + 1);
	    strncpy(skp->smk_netlabel.domain, mjstr, strlen(mjstr));

	    flag = 1;//goto the function whitch can init domain map.

	  }else{
	   if(smlen == 3){//if label len are 21
	    common_sec = secmack_parser(smack);
	    if (common_sec == NULL){
		//printk("common_sec get fail.\n");//调试信息
		goto freeout;
	    }

	    typehigh = (smack[0] - '0')&0x3;
	    pr_smkdebug("smack[1] [%c]  typehigh [%d]\n", smack[1], typehigh); //调试信息
	    //leftstr=strstr(smack, "81"); //User -> process = 0x82 -> 0x81,现已取消这个功能

	    //if(smack[1] == '4' && typehigh == 0){//已经不需要进行类型判断
		//只有用户Label转为Process密标的才走这个分支
		sprintf(temp, "%02x", common_sec->commonSM->level);
		pr_smkdebug("ver_lvl: [%02x] \n", common_sec->commonSM->level);//调试信息
		strcat(mjstr, temp);
		sprintf(temp, "%08x", common_sec->commonSM->semanteme);
		strcat(mjstr, temp);
		sprintf(temp, "%08x", common_sec->commonSM->cat);
		strcat(mjstr, temp);

		sprintf(temp, "%02x", common_sec->commonSM->self_defined);
		strcat(mjstr, temp);
		pr_smkdebug("mjstr [%s] strlen-mjstr [%lu] level [%02x] cat [%08x] smack [%s] . \n", mjstr, strlen(mjstr), common_sec->commonSM->level, common_sec->commonSM->cat, smack);//调试信息

		mjstr[21] = '\0';
		skp->smk_netlabel.domain = kzalloc(strlen(mjstr) + 1, GFP_KERNEL);
		memset(skp->smk_netlabel.domain, 0, strlen(mjstr) + 1);
		strncpy(skp->smk_netlabel.domain, mjstr, strlen(mjstr));

		flag = 1;
	    //}
	    //kfree(common_sec);
	  }
	}

        pr_smkdebug("skp->smk_netlabel.domain [%s] \n", skp->smk_netlabel.domain);//调试信息
	skp->smk_netlabel.flags =
		NETLBL_SECATTR_DOMAIN | NETLBL_SECATTR_MLS_LVL;
	/*
	 * If direct labeling works use it.
	 * Otherwise use mapped labeling.
	 */

	/*if (slen < SMK_CIPSOLEN)
		rc = smk_netlbl_mls(smack_cipso_direct, skp->smk_known,
			       &skp->smk_netlabel, slen);
	else
		rc = smk_netlbl_mls(smack_cipso_mapped, (char *)&skp->smk_secid,
			       &skp->smk_netlabel, sizeof(skp->smk_secid));
	*/  //原函数操作，现已不用，保留作为参考

	if(flag == 1){
		rc = smk_netlbl_mls(smack_cipso_direct, skp->smk_netlabel.domain,
			       &skp->smk_netlabel, strlen(skp->smk_netlabel.domain));
	}else{
	    skp->smk_netlabel.domain = skp->smk_known;
	    if (slen < SMK_CIPSOLEN)
		rc = smk_netlbl_mls(smack_cipso_direct, skp->smk_known,
			       &skp->smk_netlabel, slen);
	    else
		rc = smk_netlbl_mls(smack_cipso_mapped, (char *)&skp->smk_secid,
			       &skp->smk_netlabel, sizeof(skp->smk_secid));
	}

	if (rc >= 0) {
		INIT_LIST_HEAD(&skp->smk_rules);
		mutex_init(&skp->smk_rules_lock);
		/*
		 * Make sure that the entry is actually
		 * filled before putting it on the list.
		 */
		smk_insert_entry(skp);
		goto unlockout;
	}
	/*
	 * smk_netlbl_mls failed.
	 */
	kfree(skp->smk_netlabel.domain);
	kfree(skp);
	skp = ERR_PTR(rc);
freeout:
	kfree(smack);
unlockout:
	mutex_unlock(&smack_known_lock);
	kfree(common_sec);
	return skp;
}

/**
 * smack_from_secid - find the Smack label associated with a secid
 * @secid: an integer that might be associated with a Smack label
 *
 * Returns a pointer to the appropriate Smack label entry if there is one,
 * otherwise a pointer to the invalid Smack label.
 */
struct smack_known *smack_from_secid(const u32 secid)
{
	struct smack_known *skp;

	rcu_read_lock();
	list_for_each_entry_rcu(skp, &smack_known_list, list) {
		if (skp->smk_secid == secid) {
			rcu_read_unlock();
			return skp;
		}
	}

	/*
	 * If we got this far someone asked for the translation
	 * of a secid that is not on the list.
	 */
	rcu_read_unlock();
	return &smack_known_invalid;
}

/*
 * Unless a process is running with one of these labels
 * even having CAP_MAC_OVERRIDE isn't enough to grant
 * privilege to violate MAC policy. If no labels are
 * designated (the empty list case) capabilities apply to
 * everyone.
 */
LIST_HEAD(smack_onlycap_list);
DEFINE_MUTEX(smack_onlycap_lock);

/**
 * smack_privileged_cred - are all privilege requirements met by cred
 * @cap: The requested capability
 * @cred: the credential to use
 *
 * Is the task privileged and allowed to be privileged
 * by the onlycap rule.
 *
 * Returns true if the task is allowed to be privileged, false if it's not.
 */
bool smack_privileged_cred(int cap, const struct cred *cred)
{
	struct task_smack *tsp = cred->security;
	struct smack_known *skp = tsp->smk_task;
	struct smack_known_list_elem *sklep;
	int rc;

	//rc = cap_capable(cred, &init_user_ns, cap, CAP_OPT_NONE);//4.19.113内核版本代码，谨慎期间改用之前版本代码
	rc = cap_capable(cred, &init_user_ns, cap, 1);
	if (rc)
		return false;

	rcu_read_lock();
	if (list_empty(&smack_onlycap_list)) {
		rcu_read_unlock();
		return true;
	}

	list_for_each_entry_rcu(sklep, &smack_onlycap_list, list) {
		if (sklep->smk_label == skp) {
			rcu_read_unlock();
			return true;
		}
	}
	rcu_read_unlock();

	return false;
}

/**
 * smack_privileged - are all privilege requirements met
 * @cap: The requested capability
 *
 * Is the task privileged and allowed to be privileged
 * by the onlycap rule.
 *
 * Returns true if the task is allowed to be privileged, false if it's not.
 */
bool smack_privileged(int cap)
{
	/*
	 * All kernel tasks are privileged
	 */
	if (unlikely(current->flags & PF_KTHREAD))
		return true;

	return smack_privileged_cred(cap, current_cred());
}
